#include "vtkUnstructuredGridQuadricDecimation.h"

#include "vtkUnstructuredGrid.h"
#include "vtkPoints.h"
#include "vtkDataSetAttributes.h"
#include "vtkPointData.h"
#include "vtkCellArray.h"
#include "vtkDoubleArray.h"
#include "vtkObjectFactory.h"
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include <vtkstd/map>

class vtkUnstructuredGridQuadricDecimationEdge;
class vtkUnstructuredGridQuadricDecimationFace;
class vtkUnstructuredGridQuadricDecimationFaceHash;
class vtkUnstructuredGridQuadricDecimationFaceMap;
class vtkUnstructuredGridQuadricDecimationVec4;
class vtkUnstructuredGridQuadricDecimationSymMat4;
class vtkUnstructuredGridQuadricDecimationQEF;
class vtkUnstructuredGridQuadricDecimationTetra;
class vtkUnstructuredGridQuadricDecimationVertex;
class vtkUnstructuredGridQuadricDecimationTetMesh;


// floating point epsilons
#define VTK_FEPS 1e-6
#define VTK_TEPS 1e-6
#define VTK_SWAP(a, b, type) \
  { type t = a; a = b; b = t; }
#define VTK_PRECHECK(pointer) \
  if (pointer) \
  delete [] (pointer); \
  (pointer)


// ==================================================================================
// Vector 4 class
class vtkUnstructuredGridQuadricDecimationVec4
  {
  public:

    vtkUnstructuredGridQuadricDecimationVec4()
      {
      memset(values, 0, sizeof(values));
      }

    vtkUnstructuredGridQuadricDecimationVec4(float v1, float v2, float v3, float v4)
      {
      values[0] = v1;
      values[1] = v2;
      values[2] = v3;
      values[3] = v4;
      }

    float & operator[](const int &index);
    float operator[](const int &index) const;

    vtkUnstructuredGridQuadricDecimationVec4 operator+(const vtkUnstructuredGridQuadricDecimationVec4 &v) const;
    vtkUnstructuredGridQuadricDecimationVec4 operator-(const vtkUnstructuredGridQuadricDecimationVec4 &v) const;
    vtkUnstructuredGridQuadricDecimationVec4 operator*(const float &f) const;
    float operator*(const vtkUnstructuredGridQuadricDecimationVec4 &v) const;
    vtkUnstructuredGridQuadricDecimationVec4 operator/(const float &f) const;
    const vtkUnstructuredGridQuadricDecimationVec4& operator*=(const float &f);
    const vtkUnstructuredGridQuadricDecimationVec4& operator/=(const float &f);
    const vtkUnstructuredGridQuadricDecimationVec4& operator+=(const vtkUnstructuredGridQuadricDecimationVec4 &v);
    const vtkUnstructuredGridQuadricDecimationVec4& operator-=(const vtkUnstructuredGridQuadricDecimationVec4 &v);
    //  vtkUnstructuredGridQuadricDecimationVec4 operator*(const vtkUnstructuredGridQuadricDecimationVec4 &v);	// cross product

    // dot product
    float dot(const vtkUnstructuredGridQuadricDecimationVec4 &v) const;

    // A = e * eT
    vtkUnstructuredGridQuadricDecimationSymMat4 multTransposeSym() const;

    float length() const;
    void normalize();

    float values[4];

  private:
  };

// ==================================================================================
// Symmetric 4x4 Matrix class
// Storing lower half
class vtkUnstructuredGridQuadricDecimationSymMat4
  {
  public:

    void Identity()
      {
      memset(values, 0, sizeof(values));
      values[0] = values[2] = values[5] = values[9] = 1.;
      }

    vtkUnstructuredGridQuadricDecimationSymMat4 operator+(const vtkUnstructuredGridQuadricDecimationSymMat4 &m) const;
    vtkUnstructuredGridQuadricDecimationSymMat4 operator-(const vtkUnstructuredGridQuadricDecimationSymMat4 &m) const;
    vtkUnstructuredGridQuadricDecimationVec4 operator*(const vtkUnstructuredGridQuadricDecimationVec4 &v) const;
    vtkUnstructuredGridQuadricDecimationSymMat4 operator*(const float &f) const;
    vtkUnstructuredGridQuadricDecimationSymMat4 operator/(const float &f) const;
    const vtkUnstructuredGridQuadricDecimationSymMat4& operator*=(const float &f);
    const vtkUnstructuredGridQuadricDecimationSymMat4& operator/=(const float &f);
    const vtkUnstructuredGridQuadricDecimationSymMat4& operator+=(const vtkUnstructuredGridQuadricDecimationSymMat4 &m);
    const vtkUnstructuredGridQuadricDecimationSymMat4& operator-=(const vtkUnstructuredGridQuadricDecimationSymMat4 &m);

    float square(const vtkUnstructuredGridQuadricDecimationVec4 &v) const;

    bool ConjugateR(const vtkUnstructuredGridQuadricDecimationSymMat4 &A1, const vtkUnstructuredGridQuadricDecimationSymMat4 &A2, const vtkUnstructuredGridQuadricDecimationVec4 &p1, vtkUnstructuredGridQuadricDecimationVec4 &x) const;
    float values[10];

  private:
  };

// ==================================================================================
// vtkUnstructuredGridQuadricDecimationVec4 inline functions
inline float &vtkUnstructuredGridQuadricDecimationVec4::operator[](const int &index)
  {
  return values[index];
  }

inline float vtkUnstructuredGridQuadricDecimationVec4::operator[](const int &index) const
  {
  return values[index];
  }

inline vtkUnstructuredGridQuadricDecimationVec4 vtkUnstructuredGridQuadricDecimationVec4::operator+(const vtkUnstructuredGridQuadricDecimationVec4 &v) const
  {
  return vtkUnstructuredGridQuadricDecimationVec4(values[0] + v.values[0],
    values[1] + v.values[1],
    values[2] + v.values[2],
    values[3] + v.values[3]);
  }

inline vtkUnstructuredGridQuadricDecimationVec4 vtkUnstructuredGridQuadricDecimationVec4::operator-(const vtkUnstructuredGridQuadricDecimationVec4 &v) const
  {
  return vtkUnstructuredGridQuadricDecimationVec4(values[0] - v.values[0],
    values[1] - v.values[1],
    values[2] - v.values[2],
    values[3] - v.values[3]);
  }

inline vtkUnstructuredGridQuadricDecimationVec4 vtkUnstructuredGridQuadricDecimationVec4::operator*(const float &f) const
  {
  return vtkUnstructuredGridQuadricDecimationVec4(values[0] * f,
    values[1] * f,
    values[2] * f,
    values[3] * f);
  }

inline vtkUnstructuredGridQuadricDecimationVec4 vtkUnstructuredGridQuadricDecimationVec4::operator/(const float &f) const
  {
  return vtkUnstructuredGridQuadricDecimationVec4(values[0] / f,
    values[1] / f,
    values[2] / f,
    values[3] / f);
  }

inline float vtkUnstructuredGridQuadricDecimationVec4::operator*(const vtkUnstructuredGridQuadricDecimationVec4 &v) const
  {
  return
    values[0] * v.values[0] +
    values[1] * v.values[1] +
    values[2] * v.values[2] +
    values[3] * v.values[3];
  }

inline const vtkUnstructuredGridQuadricDecimationVec4& vtkUnstructuredGridQuadricDecimationVec4::operator*=(const float &f)
  {
  values[0] *= f;
  values[1] *= f;
  values[2] *= f;
  values[3] *= f;
  return *this;
  }

inline const vtkUnstructuredGridQuadricDecimationVec4& vtkUnstructuredGridQuadricDecimationVec4::operator/=(const float &f)
  {
  values[0] /= f;
  values[1] /= f;
  values[2] /= f;
  values[3] /= f;
  return *this;
  }

inline const vtkUnstructuredGridQuadricDecimationVec4& vtkUnstructuredGridQuadricDecimationVec4::operator+=(const vtkUnstructuredGridQuadricDecimationVec4 &v)
  {
  values[0] += v.values[0];
  values[1] += v.values[1];
  values[2] += v.values[2];
  values[3] += v.values[3];
  return *this;
  }

inline const vtkUnstructuredGridQuadricDecimationVec4& vtkUnstructuredGridQuadricDecimationVec4::operator-=(const vtkUnstructuredGridQuadricDecimationVec4 &v)
  {
  values[0] -= v.values[0];
  values[1] -= v.values[1];
  values[2] -= v.values[2];
  values[3] -= v.values[3];
  return *this;
  }

inline float vtkUnstructuredGridQuadricDecimationVec4::dot(const vtkUnstructuredGridQuadricDecimationVec4 &v) const
  {
  return
    values[0] * v.values[0] +
    values[1] * v.values[1] +
    values[2] * v.values[2] +
    values[3] * v.values[3];
  }

inline vtkUnstructuredGridQuadricDecimationSymMat4 vtkUnstructuredGridQuadricDecimationVec4::multTransposeSym() const
  {
  static vtkUnstructuredGridQuadricDecimationSymMat4 result;
  result.values[0] = values[0] * values[0];
  result.values[1] = values[0] * values[1];
  result.values[2] = values[1] * values[1];
  result.values[3] = values[0] * values[2];
  result.values[4] = values[1] * values[2];
  result.values[5] = values[2] * values[2];
  result.values[6] = values[0] * values[3];
  result.values[7] = values[1] * values[3];
  result.values[8] = values[2] * values[3];
  result.values[9] = values[3] * values[3];
  return result;
  }

inline float vtkUnstructuredGridQuadricDecimationVec4::length() const
  {
  return sqrt(values[0]*values[0] + values[1]*values[1] +
    values[2]*values[2] + values[3]*values[3]);
  }

inline void vtkUnstructuredGridQuadricDecimationVec4::normalize()
  {
  float len = length();
  if (len!=0)
    {
    values[0] /= len;
    values[1] /= len;
    values[2] /= len;
    values[3] /= len;
    }
  }
// ==================================================================================
#define SM4op(i, OP) \
  result.values[i] = values[i] OP m.values[i];
inline vtkUnstructuredGridQuadricDecimationSymMat4 vtkUnstructuredGridQuadricDecimationSymMat4::operator+(const vtkUnstructuredGridQuadricDecimationSymMat4 &m) const
  {
  static vtkUnstructuredGridQuadricDecimationSymMat4 result;
  SM4op(0,+);
  SM4op(1,+);
  SM4op(2,+);
  SM4op(3,+);
  SM4op(4,+);
  SM4op(5,+);
  SM4op(6,+);
  SM4op(7,+);
  SM4op(8,+);
  SM4op(9,+);
  return result;
  }

inline vtkUnstructuredGridQuadricDecimationSymMat4 vtkUnstructuredGridQuadricDecimationSymMat4::operator-(const vtkUnstructuredGridQuadricDecimationSymMat4 &m) const
  {
  static vtkUnstructuredGridQuadricDecimationSymMat4 result;
  SM4op(0,-);
  SM4op(1,-);
  SM4op(2,-);
  SM4op(3,-);
  SM4op(4,-);
  SM4op(5,-);
  SM4op(6,-);
  SM4op(7,-);
  SM4op(8,-);
  SM4op(9,-);
  return result;
  }
#undef SM4op

inline vtkUnstructuredGridQuadricDecimationSymMat4 vtkUnstructuredGridQuadricDecimationSymMat4::operator*(const float &f) const
  {
  vtkUnstructuredGridQuadricDecimationSymMat4 result;
  for (int i=0; i<10; i++)
    {
    result.values[i] = values[i] * f;
    }
  return result;
  }

// Matrix times Vector -> result in a Vector4
inline vtkUnstructuredGridQuadricDecimationVec4 vtkUnstructuredGridQuadricDecimationSymMat4::operator*(const vtkUnstructuredGridQuadricDecimationVec4 &v) const
  {
  return vtkUnstructuredGridQuadricDecimationVec4(values[0]*v.values[0] + values[1]*v.values[1] +
    values[3]*v.values[2] + values[6]*v.values[3],
    values[1]*v.values[0] + values[2]*v.values[1] +
    values[4]*v.values[2] + values[7]*v.values[3],
    values[3]*v.values[0] + values[4]*v.values[1] +
    values[5]*v.values[2] + values[8]*v.values[3],
    values[6]*v.values[0] + values[7]*v.values[1] +
    values[8]*v.values[2] + values[9]*v.values[3]);

  }

// vt*M*v
inline float vtkUnstructuredGridQuadricDecimationSymMat4::square(const vtkUnstructuredGridQuadricDecimationVec4 &v) const
  {
  return
    v.values[0] * (values[0]*v.values[0] + values[1]*v.values[1] +
    values[3]*v.values[2] + values[6]*v.values[3]) +
    v.values[1] * ( values[1]*v.values[0] + values[2]*v.values[1] +
    values[4]*v.values[2] + values[7]*v.values[3]) +
    v.values[2] * (values[3]*v.values[0] + values[4]*v.values[1] +
    values[5]*v.values[2] + values[8]*v.values[3]) +
    v.values[3] * (values[6]*v.values[0] + values[7]*v.values[1] +
    values[8]*v.values[2] + values[9]*v.values[3]);

  }

inline vtkUnstructuredGridQuadricDecimationSymMat4 vtkUnstructuredGridQuadricDecimationSymMat4::operator/(const float &f) const
  {
  vtkUnstructuredGridQuadricDecimationSymMat4 result;
  for (int i=0; i<10; i++)
    {
    result.values[i] = values[i] / f;
    }
  return result;
  }

inline const vtkUnstructuredGridQuadricDecimationSymMat4 & vtkUnstructuredGridQuadricDecimationSymMat4::operator*=(const float &f)
  {
  for (int i=0; i<10; i++)
    {
    values[i] *= f;
    }
  return *this;
  }

inline const vtkUnstructuredGridQuadricDecimationSymMat4 & vtkUnstructuredGridQuadricDecimationSymMat4::operator/=(const float &f)
  {
  for (int i=0; i<10; i++)
    {
    values[i] /= f;
    }
  return *this;
  }

inline const vtkUnstructuredGridQuadricDecimationSymMat4 & vtkUnstructuredGridQuadricDecimationSymMat4::operator+=(const vtkUnstructuredGridQuadricDecimationSymMat4 &m)
  {
  for (int i=0; i<10; i++)
    {
    values[i] += m.values[i];
    }
  return *this;
  }

inline const vtkUnstructuredGridQuadricDecimationSymMat4 & vtkUnstructuredGridQuadricDecimationSymMat4::operator-=(const vtkUnstructuredGridQuadricDecimationSymMat4 &m)
  {
  for (int i=0; i<10; i++)
    {
    values[i] -= m.values[i];
    }
  return *this;
  }

// ==================================================================================
// Vector 4 class implementation

bool vtkUnstructuredGridQuadricDecimationSymMat4::ConjugateR(const vtkUnstructuredGridQuadricDecimationSymMat4 &A1, const vtkUnstructuredGridQuadricDecimationSymMat4 &A2, const vtkUnstructuredGridQuadricDecimationVec4 &p1, vtkUnstructuredGridQuadricDecimationVec4 &x) const
  {
  float e(1e-3 / 4. * (values[0] + values[2] + values[5] + values[9]));
  vtkUnstructuredGridQuadricDecimationVec4 r((A1-A2)*(p1-x));
  vtkUnstructuredGridQuadricDecimationVec4 p;
  //  int k(0);
  for (int k=0; k<4; k++)
    {	
    float s(r.dot(r));
    if (s<=0) break;
    p += (r/s);
    vtkUnstructuredGridQuadricDecimationVec4 q((*this)*p);
    float t(p.dot(q));
    if (s*t<=e) break;
    r -= (q/t);
    x += (p/t);
    }
  return true;
  }

// QEF (QEF Error Function) Representation
class vtkUnstructuredGridQuadricDecimationQEF
  {
  public:
    vtkUnstructuredGridQuadricDecimationQEF(): p(0, 0, 0, 0), e(0.0)
      {}
    vtkUnstructuredGridQuadricDecimationQEF(const vtkUnstructuredGridQuadricDecimationSymMat4 &AA, const vtkUnstructuredGridQuadricDecimationVec4 &pp, const double &ee):
    A(AA), p(pp), e(ee)
      {}
    vtkUnstructuredGridQuadricDecimationQEF(const vtkUnstructuredGridQuadricDecimationQEF &Q1, const vtkUnstructuredGridQuadricDecimationQEF &Q2, const vtkUnstructuredGridQuadricDecimationVec4 &x);

    void zeros();

    void sum(const vtkUnstructuredGridQuadricDecimationQEF &Q1, const vtkUnstructuredGridQuadricDecimationQEF &Q2, const vtkUnstructuredGridQuadricDecimationVec4 &x);
    void sum(const vtkUnstructuredGridQuadricDecimationQEF &Q1, const vtkUnstructuredGridQuadricDecimationQEF &Q2);

    void scale(const double &f);

    vtkUnstructuredGridQuadricDecimationSymMat4 A;
    vtkUnstructuredGridQuadricDecimationVec4 p;
    float e;

  private:
  };

// construct Q from Q1+Q2
inline vtkUnstructuredGridQuadricDecimationQEF::vtkUnstructuredGridQuadricDecimationQEF(const vtkUnstructuredGridQuadricDecimationQEF &Q1, const vtkUnstructuredGridQuadricDecimationQEF &Q2, const vtkUnstructuredGridQuadricDecimationVec4 &x): A(Q1.A+Q2.A), p(x)
  {
  A.ConjugateR(Q1.A, Q2.A, Q1.p, p);
  e = Q1.e + Q2.e + Q1.A.square(p-Q1.p) + Q2.A.square(p-Q2.p);
  }

inline void vtkUnstructuredGridQuadricDecimationQEF::sum(const vtkUnstructuredGridQuadricDecimationQEF &Q1, const vtkUnstructuredGridQuadricDecimationQEF &Q2)
  {
  A = Q1.A + Q2.A;
  p = (Q1.p + Q2.p) * 0.5;
  A.ConjugateR(Q1.A, Q2.A, Q1.p, p);
  e = Q1.e + Q2.e + Q1.A.square(p-Q1.p) + Q2.A.square(p-Q2.p);
  }

inline void vtkUnstructuredGridQuadricDecimationQEF::sum(const vtkUnstructuredGridQuadricDecimationQEF &Q1, const vtkUnstructuredGridQuadricDecimationQEF &Q2, const vtkUnstructuredGridQuadricDecimationVec4 &x)
  {
  A = Q1.A + Q2.A;
  p = x;
  A.ConjugateR(Q1.A, Q2.A, Q1.p, p);
  e = Q1.e + Q2.e + Q1.A.square(p-Q1.p)+Q2.A.square(p-Q2.p);
  }

inline void vtkUnstructuredGridQuadricDecimationQEF::scale(const double &f)
  {
  A *= f;
  p *= f;
  e *= f;  
  }

inline void vtkUnstructuredGridQuadricDecimationQEF::zeros()
  {
  memset(A.values, 0, sizeof(A.values));
  memset(p.values, 0, sizeof(p.values));
  e = 0.0;
  }

class vtkUnstructuredGridQuadricDecimationVertex
  {
  public:
    vtkUnstructuredGridQuadricDecimationVertex(): corner(-1), fixed(0)
      {}
    vtkUnstructuredGridQuadricDecimationVertex(float ix, float iy, float iz, float is = 0.0): corner(-1), fixed(0)
      {
      Q.p = vtkUnstructuredGridQuadricDecimationVec4(ix, iy, iz, is);
      }

    vtkUnstructuredGridQuadricDecimationQEF Q;

    int corner;
	bool fixed;

  private:
  };

class vtkUnstructuredGridQuadricDecimationEdge
  {
  public:
    // NOTE: vertex list are always sorted
    vtkUnstructuredGridQuadricDecimationEdge()
      { verts[0] = verts[1] = NULL; }
    vtkUnstructuredGridQuadricDecimationEdge(vtkUnstructuredGridQuadricDecimationVertex *va, vtkUnstructuredGridQuadricDecimationVertex *vb);

    bool operator==(const vtkUnstructuredGridQuadricDecimationEdge &e) const;

    // '<' operand is useful with hash_map
    bool operator<(const vtkUnstructuredGridQuadricDecimationEdge &e) const;

    // sort the vertices to be increasing
    void sortVerts();

    void changeVerts(vtkUnstructuredGridQuadricDecimationVertex *v1, vtkUnstructuredGridQuadricDecimationVertex *v2);

    // 2 ends of the edge
    vtkUnstructuredGridQuadricDecimationVertex *verts[2];

  private:

  };

inline vtkUnstructuredGridQuadricDecimationEdge::vtkUnstructuredGridQuadricDecimationEdge(vtkUnstructuredGridQuadricDecimationVertex *va, vtkUnstructuredGridQuadricDecimationVertex *vb)
  {
  verts[0] = va;
  verts[1] = vb;
  sortVerts();
  }

// sort the verts in the increasing order
inline void vtkUnstructuredGridQuadricDecimationEdge::sortVerts()
  {
  if (verts[0]>verts[1])
    {
    vtkUnstructuredGridQuadricDecimationVertex *v = verts[1];
    verts[1] = verts[0];
    verts[0] = v;
    }
  }

inline bool vtkUnstructuredGridQuadricDecimationEdge::operator==(const vtkUnstructuredGridQuadricDecimationEdge &e) const
  {
  return (verts[0]==e.verts[0] && verts[1]==e.verts[1]);
  }

inline bool vtkUnstructuredGridQuadricDecimationEdge::operator<(const vtkUnstructuredGridQuadricDecimationEdge &e) const
  {
  if (verts[0]<e.verts[0])
    return true;
  else
    if (verts[0]==e.verts[0] && verts[1]<e.verts[1])
      return true;
  return false;
  }

inline void vtkUnstructuredGridQuadricDecimationEdge::changeVerts(vtkUnstructuredGridQuadricDecimationVertex *v1, vtkUnstructuredGridQuadricDecimationVertex *v2)
  {  
  verts[0] = v1;
  verts[1] = v2;
  sortVerts();
  }

class vtkUnstructuredGridQuadricDecimationFace
  {
  public:
    vtkUnstructuredGridQuadricDecimationFace(vtkUnstructuredGridQuadricDecimationVertex *va, vtkUnstructuredGridQuadricDecimationVertex *vb, vtkUnstructuredGridQuadricDecimationVertex *vc);

    bool operator==(const vtkUnstructuredGridQuadricDecimationFace &f) const;
    // compare by sorting the verts and one-one comparison
    bool operator<(const vtkUnstructuredGridQuadricDecimationFace &f) const;

    // has the magnitude of 2 * area of this face
    // we don't care about if it pos or neg now
    double orientation() const;

    // compute the normal of the face
    vtkUnstructuredGridQuadricDecimationVec4 normal() const;

    // check to see if a vertex belongs to this face
    bool containVertex(vtkUnstructuredGridQuadricDecimationVertex *v) const;

    // change vertex v1 on the list to vertex v2 (for edge collapsing)
    void changeVertex(vtkUnstructuredGridQuadricDecimationVertex *fromV, vtkUnstructuredGridQuadricDecimationVertex *toV);

    // sort the vertices
    void sortVerts();

    vtkUnstructuredGridQuadricDecimationVertex *verts[3];

    // find the orthonormal e1, e2, the tangent plane
    void findOrthonormal(vtkUnstructuredGridQuadricDecimationVec4 &e1, vtkUnstructuredGridQuadricDecimationVec4 &e2) const;

    // Compute the Quadric Error for this face
    void updateQuadric(float boundaryWeight = 1.);

  private:
  };

inline vtkUnstructuredGridQuadricDecimationFace::vtkUnstructuredGridQuadricDecimationFace(vtkUnstructuredGridQuadricDecimationVertex *va, vtkUnstructuredGridQuadricDecimationVertex *vb, vtkUnstructuredGridQuadricDecimationVertex *vc)
  { 
  verts[0] = va;
  verts[1] = vb;
  verts[2] = vc;
  sortVerts();
  }

inline bool vtkUnstructuredGridQuadricDecimationFace::operator==(const vtkUnstructuredGridQuadricDecimationFace &f) const
  {
  return (verts[0]==f.verts[0] && verts[1]==f.verts[1] && verts[2]==f.verts[2]);
  }

// NOTE: the vertices of the Face are always sorted!
inline bool vtkUnstructuredGridQuadricDecimationFace::operator<(const vtkUnstructuredGridQuadricDecimationFace &f) const
  {
  register int i;
  for (i = 0; i<3; i++)
    if (verts[i]<f.verts[i])
      return true;
    else
      if (verts[i]>f.verts[i])
        return false;
  return false;
  }

inline void vtkUnstructuredGridQuadricDecimationFace::sortVerts()
  {
  if ((unsigned long)verts[1]<(unsigned long)verts[0] &&
    (unsigned long)verts[1]<(unsigned long)verts[2])
    VTK_SWAP(verts[0], verts[1], vtkUnstructuredGridQuadricDecimationVertex*);

  if ((unsigned long)verts[2]<(unsigned long)verts[0] &&
    (unsigned long)verts[2]<(unsigned long)verts[1])
    VTK_SWAP(verts[0], verts[2], vtkUnstructuredGridQuadricDecimationVertex*);

  if ((unsigned long)verts[2]<(unsigned long)verts[1])
    VTK_SWAP(verts[1], verts[2], vtkUnstructuredGridQuadricDecimationVertex*);
  }

//=============================================================
// Face_hash class declaration
// bundle of support functions for Face hash_map
class vtkUnstructuredGridQuadricDecimationFaceHash
  {
  public:

    // This is the average load that the hash table tries to maintain.
    const static size_t bucket_size = 4;

    // This is the minimum number of buckets in the hash table.  It must be
    // a positive power of two.
    const static size_t min_buckets = 8;

    // This method must define an ordering on two faces.  It is used to compare
    // faces that has to the same bucket.
    bool operator() (const vtkUnstructuredGridQuadricDecimationFace &f1, const vtkUnstructuredGridQuadricDecimationFace &f2) const
      {
      return f1 < f2;
      }

    // This produces the actual hash code for the Face
    size_t operator() (const vtkUnstructuredGridQuadricDecimationFace &f) const
      {
      return (size_t)f.verts[0] *
        (size_t)f.verts[1] * (size_t)f.verts[2];
      }

  };

//=============================================================
// FaceMap declaration
// try to make a hash_map from Face -> Face *
typedef vtkstd::map<vtkUnstructuredGridQuadricDecimationFace, vtkUnstructuredGridQuadricDecimationFace *, vtkUnstructuredGridQuadricDecimationFaceHash> vtkUnstructuredGridQuadricDecimationFaceHashMap;

class vtkUnstructuredGridQuadricDecimationFaceMap
  {
  public:
    vtkUnstructuredGridQuadricDecimationFaceMap()
      {}
    ~vtkUnstructuredGridQuadricDecimationFaceMap()
      { clear(); }

    // for iteration
    vtkUnstructuredGridQuadricDecimationFaceHashMap::iterator begin()
      { return faces.begin(); }
    vtkUnstructuredGridQuadricDecimationFaceHashMap::iterator end()
      { return faces.end(); }

    // clear
    void clear();
    size_t size() const
      { return faces.size(); }

    // insert a new face
    vtkUnstructuredGridQuadricDecimationFace * addFace(const vtkUnstructuredGridQuadricDecimationFace &f);

    // return the face that is the same as f
    // if there's no such face, then return NULL
    vtkUnstructuredGridQuadricDecimationFace *getFace(const vtkUnstructuredGridQuadricDecimationFace &f);

    // remove the face has the content f
    void removeFace(const vtkUnstructuredGridQuadricDecimationFace &f);

    // add a face, and check if it can't be a border face
    // then kill it. Return NULL -> failed adding
    vtkUnstructuredGridQuadricDecimationFace * addFaceBorder(const vtkUnstructuredGridQuadricDecimationFace &f);

    vtkUnstructuredGridQuadricDecimationFaceHashMap faces;
  private:
    // add face without checking existence
    vtkUnstructuredGridQuadricDecimationFace * directAddFace(const vtkUnstructuredGridQuadricDecimationFace &f);

    // remove face without checking existence
    void directRemoveFace(vtkUnstructuredGridQuadricDecimationFace *f);
    void directRemoveFace(vtkUnstructuredGridQuadricDecimationFaceHashMap::iterator i);
  };

//=============================================================
// Face class implementation
// NOTE: the vertices of the Face are always sorted!
// construct the face, sort the vertice as well
// calculate the orientation, or 2*area of the face
double vtkUnstructuredGridQuadricDecimationFace::orientation() const
  {
  vtkUnstructuredGridQuadricDecimationVec4 v;
  // this is cross product of v1-v0 and v2-v0
  v[0] = (verts[1]->Q.p[1] - verts[0]->Q.p[1]) * (verts[2]->Q.p[2] - verts[0]->Q.p[2])
    - (verts[2]->Q.p[1] - verts[0]->Q.p[1]) * (verts[1]->Q.p[2] - verts[0]->Q.p[2]);

  v[1] = - (verts[1]->Q.p[0] - verts[0]->Q.p[0]) * (verts[2]->Q.p[2] - verts[0]->Q.p[2])
    + (verts[2]->Q.p[0] - verts[0]->Q.p[0]) * (verts[1]->Q.p[2] - verts[0]->Q.p[2]);

  v[2] = (verts[1]->Q.p[0] - verts[0]->Q.p[0]) * (verts[2]->Q.p[1] - verts[0]->Q.p[1])
    - (verts[2]->Q.p[0] - verts[0]->Q.p[0]) * (verts[1]->Q.p[1] - verts[0]->Q.p[1]);

  v[3] = 0;

  return v.length();
  }

vtkUnstructuredGridQuadricDecimationVec4 vtkUnstructuredGridQuadricDecimationFace::normal() const
  {
  vtkUnstructuredGridQuadricDecimationVec4 v;
  // this is cross product of v1-v0 and v2-v0
  v[0] = (verts[1]->Q.p[1] - verts[0]->Q.p[1]) * (verts[2]->Q.p[2] - verts[0]->Q.p[2])
    - (verts[2]->Q.p[1] - verts[0]->Q.p[1]) * (verts[1]->Q.p[2] - verts[0]->Q.p[2]);

  v[1] = - (verts[1]->Q.p[0] - verts[0]->Q.p[0]) * (verts[2]->Q.p[2] - verts[0]->Q.p[2])
    + (verts[2]->Q.p[0] - verts[0]->Q.p[0]) * (verts[1]->Q.p[2] - verts[0]->Q.p[2]);

  v[2] = (verts[1]->Q.p[0] - verts[0]->Q.p[0]) * (verts[2]->Q.p[1] - verts[0]->Q.p[1])
    - (verts[2]->Q.p[0] - verts[0]->Q.p[0]) * (verts[1]->Q.p[1] - verts[0]->Q.p[1]);

  v[3] = 0;  
  return v/v.length();
  }

// Compute the Quadric Error for this face
void vtkUnstructuredGridQuadricDecimationFace::updateQuadric(float boundaryWeight)
  {
  vtkUnstructuredGridQuadricDecimationVec4 e1, e2;

  e1 = verts[1]->Q.p - verts[0]->Q.p;
  e2 = verts[2]->Q.p - verts[0]->Q.p;

  e1.normalize();

  e2 = e2 - e1*e2.dot(e1);
  e2.normalize();

  // A = I - e1.e1T - e2.e2T
  static vtkUnstructuredGridQuadricDecimationSymMat4 A;
  A.Identity();
  A -= (e1.multTransposeSym() + e2.multTransposeSym());
  A *= fabs(orientation())/6.0 * boundaryWeight;
  //  A *= boundaryWeight;
  for (int i=0; i<3; i++)
    {
    verts[i]->Q.A += A;
    }
  }

bool vtkUnstructuredGridQuadricDecimationFace::containVertex(vtkUnstructuredGridQuadricDecimationVertex *v) const
  {
  for (int i=0; i<3; i++)
    {
    if (verts[i]==v)
      return true;
    }
  return false;
  }

// change vertex v1 on the list to vertex v2 (for edge collapsing)
void vtkUnstructuredGridQuadricDecimationFace::changeVertex(vtkUnstructuredGridQuadricDecimationVertex *v1, vtkUnstructuredGridQuadricDecimationVertex *v2)
  {
  for (int i=0; i<3; i++)
    {
    if (verts[i]==v1)
      verts[i] = v2;
    }
  sortVerts();
  }

//=============================================================
// FaceMap class implementation
void vtkUnstructuredGridQuadricDecimationFaceMap::clear()
  {
  vtkUnstructuredGridQuadricDecimationFaceHashMap::iterator i = faces.begin();
  // free all the memory
  while (i!=faces.end())
    {
    delete (*i).second;
    i++;
    }
  // clear the hash table
  faces.clear();
  }

// insert in new faces
vtkUnstructuredGridQuadricDecimationFace * vtkUnstructuredGridQuadricDecimationFaceMap::addFace(const vtkUnstructuredGridQuadricDecimationFace &f)
  {
  if (getFace(f)==NULL)
    {
    return directAddFace(f);
    } else
    {
    return NULL;
    }
  }

// return the face that is the same as f
// if there's no such face, then return NULL
vtkUnstructuredGridQuadricDecimationFace * vtkUnstructuredGridQuadricDecimationFaceMap::getFace(const vtkUnstructuredGridQuadricDecimationFace &f)
  {
  vtkUnstructuredGridQuadricDecimationFaceHashMap::iterator i = faces.find(f);
  if (i!=faces.end())
    return (*i).second;
  else
    return NULL;
  }

// remove the face has the content f
void vtkUnstructuredGridQuadricDecimationFaceMap::removeFace(const vtkUnstructuredGridQuadricDecimationFace &f)
  {
  vtkUnstructuredGridQuadricDecimationFaceHashMap::iterator i = faces.find(f);
  if (i!=faces.end())
    {
    directRemoveFace(i);
    }
  }

// add a face, and check if it can't be a border face
// then kill it. Return NULL -> failed adding
vtkUnstructuredGridQuadricDecimationFace * vtkUnstructuredGridQuadricDecimationFaceMap::addFaceBorder(const vtkUnstructuredGridQuadricDecimationFace &f)
  {
  vtkUnstructuredGridQuadricDecimationFaceHashMap::iterator i = faces.find(f);
  if (i!=faces.end())
    { // exist -> has 2 tets -> not a border -> kill it
    directRemoveFace(i);
    return NULL;
    } else
    { 	// not exist -> add it in
    return directAddFace(f);
    }
  }

// add face without checking existence
vtkUnstructuredGridQuadricDecimationFace * vtkUnstructuredGridQuadricDecimationFaceMap::directAddFace(const vtkUnstructuredGridQuadricDecimationFace &f)
  {
  vtkUnstructuredGridQuadricDecimationFace *newF = new vtkUnstructuredGridQuadricDecimationFace(f);
  faces[f] = newF;
  return newF;
  }

// remove face without checking existence
void vtkUnstructuredGridQuadricDecimationFaceMap::directRemoveFace(vtkUnstructuredGridQuadricDecimationFace *f)
  {
  faces.erase(*f);
  delete f;
  }

void vtkUnstructuredGridQuadricDecimationFaceMap::directRemoveFace(vtkUnstructuredGridQuadricDecimationFaceHashMap::iterator i)
  {
  vtkUnstructuredGridQuadricDecimationFace *f = (*i).second;
  faces.erase(i);
  delete f;
  }

class vtkUnstructuredGridQuadricDecimationTetra
  {
  public:
    vtkUnstructuredGridQuadricDecimationTetra(): index(-1)
      {}
    vtkUnstructuredGridQuadricDecimationTetra(vtkUnstructuredGridQuadricDecimationVertex *va, vtkUnstructuredGridQuadricDecimationVertex *vb, vtkUnstructuredGridQuadricDecimationVertex *vc, vtkUnstructuredGridQuadricDecimationVertex *vd);

    vtkUnstructuredGridQuadricDecimationVertex *verts[4];	// pointers to 4 vertices

    // 
    // the orientation of this order of vertices
    // positive - good orientation
    // zero - all in one plane
    // negative - bad orientation
    //
    // it is also 6 times the volume of this tetrahedra
    float orientation() const;
    float orientation(const vtkUnstructuredGridQuadricDecimationVec4 &v0,
      const vtkUnstructuredGridQuadricDecimationVec4 &v1,
      const vtkUnstructuredGridQuadricDecimationVec4 &v2,
      const vtkUnstructuredGridQuadricDecimationVec4 &v3) const;

    // swap vertices so that the orientation is positive
    void fixOrientation();

    // check to see if a vertex belongs to this tetrahedron
    bool containVertex(vtkUnstructuredGridQuadricDecimationVertex *v) const;

    // check to see if we can change fromV to toV without changing the orietation
    bool changable(vtkUnstructuredGridQuadricDecimationVertex *fromV, const vtkUnstructuredGridQuadricDecimationVec4 &v4);

    // change vertex v1 on the list to vertex v2 (for edge collapsing)
    void changeVertex(vtkUnstructuredGridQuadricDecimationVertex *fromV, vtkUnstructuredGridQuadricDecimationVertex *toV);

    // find the orthonormal e1, e2, e3, the tangent space
    void findOrthonormal(vtkUnstructuredGridQuadricDecimationVec4 &e1, vtkUnstructuredGridQuadricDecimationVec4 &e2, vtkUnstructuredGridQuadricDecimationVec4 &e3) const;

    // Compute the Quadric Error for this tetrahedron
    void updateQuadric();

    int index;
  private:
  };

inline vtkUnstructuredGridQuadricDecimationTetra::vtkUnstructuredGridQuadricDecimationTetra(vtkUnstructuredGridQuadricDecimationVertex *va, vtkUnstructuredGridQuadricDecimationVertex *vb, vtkUnstructuredGridQuadricDecimationVertex *vc, vtkUnstructuredGridQuadricDecimationVertex *vd)
  {
  verts[0] = va;
  verts[1] = vb;
  verts[2] = vc;
  verts[3] = vd;
  }

inline bool vtkUnstructuredGridQuadricDecimationTetra::containVertex(vtkUnstructuredGridQuadricDecimationVertex *v) const
  {
  if (verts[0]==v ||
    verts[1]==v ||
    verts[2]==v ||
    verts[3]==v)
    return true;
  return false;
  }

#define U(c) (verts[1]->Q.p[c] - verts[0]->Q.p[c])
#define V(c) (verts[2]->Q.p[c] - verts[0]->Q.p[c])
#define W(c) (verts[3]->Q.p[c] - verts[0]->Q.p[c])
inline float vtkUnstructuredGridQuadricDecimationTetra::orientation() const
  {
  return
    U(0) * (V(1)*W(2) - V(2)*W(1)) -
    V(0) * (U(1)*W(2) - U(2)*W(1)) +
    W(0) * (U(1)*V(2) - U(2)*V(1));
  }
#undef U
#undef V
#undef W

#define U(c) (v1[c] - v0[c])
#define V(c) (v2[c] - v0[c])
#define W(c) (v3[c] - v0[c])
inline float vtkUnstructuredGridQuadricDecimationTetra::orientation(const vtkUnstructuredGridQuadricDecimationVec4 &v0,
                                                                    const vtkUnstructuredGridQuadricDecimationVec4 &v1,
                                                                    const vtkUnstructuredGridQuadricDecimationVec4 &v2,
                                                                    const vtkUnstructuredGridQuadricDecimationVec4 &v3) const
  {
  return
    U(0) * (V(1)*W(2) - V(2)*W(1)) -
    V(0) * (U(1)*W(2) - U(2)*W(1)) +
    W(0) * (U(1)*V(2) - U(2)*V(1));
  }
#undef U
#undef V
#undef W

// check to see if we can change fromV to toV without changing the orietation
inline bool vtkUnstructuredGridQuadricDecimationTetra::changable(vtkUnstructuredGridQuadricDecimationVertex *fromV, const vtkUnstructuredGridQuadricDecimationVec4 &v4)
  {
  if (fromV==verts[0])
    return orientation(v4, verts[1]->Q.p, verts[2]->Q.p, verts[3]->Q.p)>VTK_TEPS;
  if (fromV==verts[1])
    return orientation(verts[0]->Q.p, v4, verts[2]->Q.p, verts[3]->Q.p)>VTK_TEPS;
  if (fromV==verts[2])
    return orientation(verts[0]->Q.p, verts[1]->Q.p, v4, verts[3]->Q.p)>VTK_TEPS;
  if (fromV==verts[3])
    return orientation(verts[0]->Q.p, verts[1]->Q.p, verts[2]->Q.p, v4)>VTK_TEPS;
  return true; 
  }

void vtkUnstructuredGridQuadricDecimationTetra::fixOrientation()
  {
  if (orientation()<0)
    {
    VTK_SWAP(verts[2], verts[3], vtkUnstructuredGridQuadricDecimationVertex *);
    }
  if (orientation()<0)
    {
    VTK_SWAP(verts[1], verts[2], vtkUnstructuredGridQuadricDecimationVertex *);
    }
  }

// change vertex v1 on the list to vertex v2 (for edge collapsing)
void vtkUnstructuredGridQuadricDecimationTetra::changeVertex(vtkUnstructuredGridQuadricDecimationVertex *v1, vtkUnstructuredGridQuadricDecimationVertex *v2)
  {
  for (int i=0; i<4; i++)
    {
    if (verts[i]==v1)
      verts[i] = v2;
    }
  }

// find the orthonormal tangent space e1, e2, e3
void vtkUnstructuredGridQuadricDecimationTetra::findOrthonormal(vtkUnstructuredGridQuadricDecimationVec4 &e1, vtkUnstructuredGridQuadricDecimationVec4 &e2, vtkUnstructuredGridQuadricDecimationVec4 &e3) const
  {
  vtkUnstructuredGridQuadricDecimationVec4 e0(verts[0]->Q.p);

  // Ei = Ui - U0
  e1 = verts[1]->Q.p - e0;
  e2 = verts[2]->Q.p - e0;
  e3 = verts[3]->Q.p - e0;

  e1.normalize();

  e2 = e2 - e1*e2.dot(e1);
  e2.normalize();

  e3 = e3 - e1*e3.dot(e1) - e2*e3.dot(e2);
  e3.normalize();
  }

// compute the quadric error based on this tet
#define ax a.values[0]
#define ay a.values[1]
#define az a.values[2]
#define af a.values[3]
#define bx b.values[0]
#define by b.values[1]
#define bz b.values[2]
#define bf b.values[3]
#define cx c.values[0]
#define cy c.values[1]
#define cz c.values[2]
#define cf c.values[3]
void vtkUnstructuredGridQuadricDecimationTetra::updateQuadric()
  {
  vtkUnstructuredGridQuadricDecimationVec4 a(verts[1]->Q.p - verts[0]->Q.p);
  vtkUnstructuredGridQuadricDecimationVec4 b(verts[2]->Q.p - verts[0]->Q.p);
  vtkUnstructuredGridQuadricDecimationVec4 c(verts[3]->Q.p - verts[0]->Q.p);
  vtkUnstructuredGridQuadricDecimationVec4 n(ay*(bz*cf-bf*cz) + az*(bf*cy-by*cf) + af*(by*cz-bz*cy),
    az*(bx*cf-bf*cx) + af*(bz*cx-bx*cz) + ax*(bf*cz-bz*cf),
    af*(bx*cy-by*cx) + ax*(by*cf-bf*cy) + ay*(bf*cx-bx*cf),
    ax*(bz*cy-by*cz) + ay*(bx*cz-bz*cx) + az*(by*cx-bx*cy));
  vtkUnstructuredGridQuadricDecimationSymMat4 A(n.multTransposeSym());
  // weight by the volume of the tet
  //  Q.scale(orientation()/6.0);
  // we want to divide by 4 also, for each vertex
  A *= 1.5 / fabs(orientation());

  for (int i=0; i<4; i++)
    {
    verts[i]->Q.A += A;
    }
  }
#undef ax
#undef ay
#undef az
#undef af
#undef bx
#undef by
#undef bz
#undef bf
#undef cx
#undef cy
#undef cz
#undef cf

class vtkUnstructuredGridQuadricDecimationTetMesh
  {
  public:
    vtkUnstructuredGridQuadricDecimationTetMesh(): setSize(8), doublingRatio(0.4), noDoubling(false),
      boundaryWeight(100.0), 
      verts(NULL), tets(NULL), PT(NULL), 
      unusedTets(0), unusedVerts(0), L(NULL)
      {
      }

    ~vtkUnstructuredGridQuadricDecimationTetMesh()
      {
      clear();
      }

    void addTet(vtkUnstructuredGridQuadricDecimationTetra *t);

    void clear();	// clear the mesh -> empty

    // OFF file manipulator
    bool loadVtkUnstructuredGrid(vtkUnstructuredGrid *vgrid);
    bool saveVtkUnstructuredGrid(vtkUnstructuredGrid *vgrid);

    // Simplification
    int setSize;
    float doublingRatio;
    bool noDoubling;
    float boundaryWeight;
    void buildFullMesh();
    int simplify(int n, int desiredTets);

    int vCount;
    int tCount;
    vtkUnstructuredGridQuadricDecimationVertex *verts;
    vtkUnstructuredGridQuadricDecimationTetra *tets;
    vtkUnstructuredGridQuadricDecimationTetra **PT;
    vtkUnstructuredGridQuadricDecimationFaceMap faces;

    // number of tets deleted but not free
    int unusedTets;
    int unusedVerts;
    int maxTet;

    int *L;

  private:
    void addCorner(vtkUnstructuredGridQuadricDecimationVertex *v, int corner);
    void removeCorner(vtkUnstructuredGridQuadricDecimationVertex *v, int corner);
    void removeVertex(vtkUnstructuredGridQuadricDecimationVertex *v);  

    // really clean up the memory of remove_tet
    void doDelete();

    // check if this edge can be collapsed (i.e. without violating boundary, vol...)
    bool contractable(vtkUnstructuredGridQuadricDecimationEdge &e, const vtkUnstructuredGridQuadricDecimationVec4 &target);

    // Simplification
    void mergeTets(vtkUnstructuredGridQuadricDecimationVertex *dst, vtkUnstructuredGridQuadricDecimationVertex *src);
    void deleteMin (int setSize, vtkUnstructuredGridQuadricDecimationEdge &e, vtkUnstructuredGridQuadricDecimationQEF &Q);
  };

#define VTK_ADDFACEBORDER(i,j,k) \
  faces.addFaceBorder(vtkUnstructuredGridQuadricDecimationFace(t->verts[i], t->verts[j], t->verts[k]))
void vtkUnstructuredGridQuadricDecimationTetMesh::addTet(vtkUnstructuredGridQuadricDecimationTetra *t)
  {
  if (t->orientation()<-VTK_FEPS)
    t->fixOrientation();

  // add all of its faces to the FaceMap => 4 faces
  // NOTE: adding faces to vertices' list will be done
  // after we have all the faces (because some faces
  // might be deleted if it is not on the surface!!!)
  VTK_ADDFACEBORDER(0,1,2);
  VTK_ADDFACEBORDER(0,1,3);
  VTK_ADDFACEBORDER(0,2,3);
  VTK_ADDFACEBORDER(1,2,3);
  }
#undef VTK_ADDFACEBORDER

void vtkUnstructuredGridQuadricDecimationTetMesh::removeVertex(vtkUnstructuredGridQuadricDecimationVertex *v)
  {
  v->corner = -1;
  unusedVerts++;
  }

void vtkUnstructuredGridQuadricDecimationTetMesh::addCorner(vtkUnstructuredGridQuadricDecimationVertex *v, int corner)
  {
  if (v->corner<0)
    {
    v->corner = corner;
    L[corner] = corner;	
    }
  else
    {
    L[corner] = L[v->corner];
    L[v->corner] = corner;
    }
  }

void vtkUnstructuredGridQuadricDecimationTetMesh::removeCorner(vtkUnstructuredGridQuadricDecimationVertex *v, int base)
  {  
  int c = v->corner;  
  if (c/4==base)
    {
    while (L[c]!=v->corner)
      {
      c = L[c];
      }
    L[c] = L[v->corner];
    v->corner = c;
    }
  else
    {
    do
      {	  
      if (L[c]/4==base)
        {
        L[c] = L[L[c]];
        return;
        }
      c = L[c];
      } while (c!=v->corner);
    }
  }

// really clean up the memory of remove_tet
void vtkUnstructuredGridQuadricDecimationTetMesh::doDelete()
  {
  int i = 0;
  int j = tCount-1;
  do
    {
    while (i<=j && tets[i].index>=0) i++;
    while (j>=i && tets[j].index<0) j--;
    if (i<j)
      {
      tets[i] = tets[j];
      i++; j--;
      }
    } while (i<=j);
    //  tets.resize(i);
    unusedTets = 0;

    i = 0;
    j = vCount-1;
    do
      {
      while (i<=j && verts[i].corner>=0) i++;
      while (j>=i && verts[j].corner<0) j--;
      if (i<j)
        {
        verts[i] = verts[j];
        i++; j--;
        }
      } while (i<=j);
      //  verts.resize(i);
      unusedVerts = 0;
  }

// Clean the mesh
void vtkUnstructuredGridQuadricDecimationTetMesh::clear()
  {
  VTK_PRECHECK(verts) = NULL;
  VTK_PRECHECK(tets) = NULL;
  VTK_PRECHECK(PT) = NULL;
  VTK_PRECHECK(L) = NULL;
  faces.clear();
  unusedTets = 0;
  unusedVerts = 0;
  }

// SIMPLIFICATION IMPLEMNETATION

// buildFullMesh
//	- Adding faces to vertices list and initialize theirs quadrics
//	- Compute quadric error at each vertex or remove unused vertices
void vtkUnstructuredGridQuadricDecimationTetMesh::buildFullMesh()
  {
  vtkUnstructuredGridQuadricDecimationFaceHashMap::iterator fi = faces.begin();
  while (fi!=faces.end())
    {
    vtkUnstructuredGridQuadricDecimationFace *f = (*fi).second;
    f->updateQuadric(boundaryWeight);
    fi++;
    }
  }

void vtkUnstructuredGridQuadricDecimationTetMesh::deleteMin (int setSize, vtkUnstructuredGridQuadricDecimationEdge &finalE, vtkUnstructuredGridQuadricDecimationQEF &minQ)
  {  
  // Multiple Choice Randomize set
  static float lasterror = 0;
  bool stored(false);
  vtkUnstructuredGridQuadricDecimationQEF Q;
  vtkUnstructuredGridQuadricDecimationEdge e(NULL, NULL);
  for (int j=0; j<2; j++)
    {
    for (int i=0; i<setSize; i++)
      {
      int k = rand() % maxTet;
      if (tets[k].index<0)
        {
        do
          {
          maxTet--;
          } while (maxTet>0 && tets[maxTet].index<0);	  
          if (k<maxTet)
            {
            int idx = tets[k].index;
            tets[k] = tets[maxTet];
            tets[maxTet].index = idx;
            PT[tets[k].index] = &tets[k];
            }
          else
            k = maxTet++;
        }	

      e.verts[0] = tets[k].verts[rand() % 4];
      do
        {
        e.verts[1] = tets[k].verts[rand() % 4];
        } while (e.verts[1]==e.verts[0]);

        if (!stored)
          {
          finalE = e;
          minQ.sum(e.verts[0]->Q, e.verts[1]->Q);
          stored = true;
          }
        else
          {
          if (e.verts[0]->Q.e + e.verts[1]->Q.e < minQ.e)
            {
            Q.sum(e.verts[0]->Q, e.verts[1]->Q);
            if (Q.e<minQ.e)
              {
              finalE = e;
              minQ = Q;
              }
            }
          }
      }
    if (noDoubling || (minQ.e-lasterror)/lasterror<=doublingRatio) break;
    }
  lasterror = minQ.e;
  }

// Simplify the mesh by a series of N edge contractions
// or to the number of desiredTets
// it returns the actual number of edge contractions
int vtkUnstructuredGridQuadricDecimationTetMesh::simplify(int n, int desiredTets)
  {
  int count = 0;
  int uncontractable = 0;
  int run = 0;
  while ((count<n || desiredTets<(tCount-unusedTets)) &&
    (run<1000))
    {	// as long as we want to collapse

    vtkUnstructuredGridQuadricDecimationQEF Q;
    vtkUnstructuredGridQuadricDecimationEdge e;

    deleteMin(setSize, e, Q);

    if (contractable(e, Q.p))
      {
      run = 0;
      // begin to collapse the edge Va + Vb -> Va = e.target
      vtkUnstructuredGridQuadricDecimationVertex *va = e.verts[0];
      vtkUnstructuredGridQuadricDecimationVertex *vb = e.verts[1];

      // Constructing new vertex
      va->Q = Q;

      // Merge all faces and tets of Va and Vb, remove the degenerated ones
      mergeTets(va, vb);
      vb->corner = -1;
      unusedVerts++;

      // Complete the edge contraction
      count++;
      } else 
      {
      uncontractable++;
      run++;
      }
    }
  return count;
  }

// Merge all tets of Vb to Va by changing Vb to Va
// and add all tets of Vb to Va's Tet List
// Also, it will remove all tets contain both Vb and Va
// In fact, this is merging corners
void vtkUnstructuredGridQuadricDecimationTetMesh::mergeTets(vtkUnstructuredGridQuadricDecimationVertex *dst, vtkUnstructuredGridQuadricDecimationVertex *src)
  {
  int next = src->corner;
  vtkUnstructuredGridQuadricDecimationTetra *t = NULL;
  do
    {
    t = PT[next/4];
    if (t)
      {
      if (t->containVertex(dst))
        {
        t->index = -t->index-1;
        unusedTets++;
        PT[next/4] = NULL;
        }
      else
        {
        t->changeVertex(src, dst);
        }
      }
    next = L[next];
    } while (next!=src->corner);

    // Then we merge them all together
    next = L[dst->corner];
    L[dst->corner] = L[src->corner];
    L[src->corner] = next;
    bool notstop = true;
    int prev = dst->corner;
    next = L[prev];
    do
      {
      notstop = next!=dst->corner;
      t = PT[next/4];
      if (!t)
        {
        next = L[next];
        L[prev] = next;
        }
      else
        {
        prev = next;
        next = L[next];
        }
      } while (notstop);
      dst->corner = prev;
  }

// Check if an edge can be contracted
bool vtkUnstructuredGridQuadricDecimationTetMesh::contractable(vtkUnstructuredGridQuadricDecimationEdge &e, const vtkUnstructuredGridQuadricDecimationVec4 &target)
  {
  // need to check all the tets around both vertices to see if they can
  // adapt the new target vertex or not
  if (e.verts[0]->fixed || e.verts[1]->fixed)
	return false;
  vtkUnstructuredGridQuadricDecimationTetra *t = NULL;
  for (int i=0; i<2; i++)
    {
    int c = e.verts[i]->corner;
    do
      {
      t = PT[c/4];
      if (t)
        if (!(t->containVertex(e.verts[0]) &&
          t->containVertex(e.verts[1])) &&
          !(t->changable(e.verts[i], target)))
          {
          return false;
          }
        c = L[c];
      } while (c!=e.verts[i]->corner);
    }
  return true;
  }

bool vtkUnstructuredGridQuadricDecimationTetMesh::loadVtkUnstructuredGrid(vtkUnstructuredGrid *vgrid)
  {
  clear();
  // Read all the vertices first
  vCount = vgrid->GetNumberOfPoints();
  VTK_PRECHECK(verts) = new vtkUnstructuredGridQuadricDecimationVertex[vCount];
  vtkPoints *vp = vgrid->GetPoints();
  //  vtkDataArray *vs = vgrid->GetPointData()->GetAttribute(vtkDataSetAttributes::SCALARS);
  vtkDataArray *vs = vgrid->GetPointData()->GetArray("scalars");
  vtkDataArray *finals = vgrid->GetPointData()->GetArray("final");
  for (int i=0; i<vCount; i++)
    {
    double *pos = vp->GetPoint(i);
    double *scalar = vs->GetTuple(i);
	if (finals) {
		double *final = finals->GetTuple(i);
		verts[i].fixed = (*final)<0.5;
	}
	else
		verts[i].fixed = false;
    verts[i].Q.p[0] = pos[0];
    verts[i].Q.p[1] = pos[1];
    verts[i].Q.p[2] = pos[2];	
    verts[i].Q.p[3] = scalar[0];
    }

  // Read all the tets
  tCount = vgrid->GetNumberOfCells();
  maxTet = tCount;
  VTK_PRECHECK(tets) = new vtkUnstructuredGridQuadricDecimationTetra[tCount];
  VTK_PRECHECK(PT) = new (vtkUnstructuredGridQuadricDecimationTetra(*[tCount]));
  VTK_PRECHECK(L) = new int[4 * tCount];
  vtkCellArray *vt = vgrid->GetCells();
  vtkIdType npts;
  vtkIdType *idx;
  vtkIdType curIdx = 0;
  for (int i=0; i<tCount; i++)
    {
    vt->GetCell(curIdx, npts, idx);
    curIdx += (npts+1);
    if (npts==4)
      {
      for (int k=0; k<4; k++)
        {
        tets[i].verts[k] = &verts[idx[k]];
        }
      addTet(&tets[i]);
      addCorner(tets[i].verts[0], i*4 + 0);
      addCorner(tets[i].verts[1], i*4 + 1);
      addCorner(tets[i].verts[2], i*4 + 2);
      addCorner(tets[i].verts[3], i*4 + 3);	
      tets[i].updateQuadric();
      PT[i] = &tets[i];
      tets[i].index = i;
      }
    }

  return true;
  }

bool vtkUnstructuredGridQuadricDecimationTetMesh::saveVtkUnstructuredGrid(vtkUnstructuredGrid *vgrid)
  {
  vtkIdType growSize = (tCount-unusedTets)*4;	
  vgrid->Allocate(growSize, growSize);
  vtkPoints *vp = vtkPoints::New();
  vtkDoubleArray *vs = vtkDoubleArray::New();

  // Output vertices
  // We need a map for indexing
  vtkstd::map<vtkUnstructuredGridQuadricDecimationVertex *, int> indexes;
  vp->SetNumberOfPoints(vCount-unusedVerts);
  vs->SetNumberOfValues(vCount-unusedVerts);
  int vIdx = 0;
  for (int i=0; i<vCount; i++)
    {
    if (verts[i].corner>=0)
      {
      vp->SetPoint(vIdx, verts[i].Q.p[0], verts[i].Q.p[1], verts[i].Q.p[2]);
      vs->SetValue(vIdx, verts[i].Q.p[3]);
      indexes[&verts[i]] = vIdx++;
      }
    }
  vgrid->SetPoints(vp);
  vp->Delete();
  vs->SetName("scalars");
  vgrid->GetPointData()->AddArray(vs);
  vgrid->GetPointData()->SetScalars(vs);
  vs->Delete();

  vtkIdType idx[4];
  for (int i=0; i<maxTet; i++)
    {
    if (tets[i].index>=0)
      {
      for (int j=0; j<4; j++)
        idx[j] = indexes[tets[i].verts[j]];
      vgrid->InsertNextCell(VTK_TETRA, 4, idx);
      }
    }
  return true;
  }

#undef VTK_PRECHECK
#undef VTK_FEPS
#undef VTK_TEPS
#undef VTK_SWAP


////////////////////////////////////////////////////////////////////////////////
/* ========================================================================== */
////////////////////////////////////////////////////////////////////////////////
/* === vtkUnstructuredGridQuadricDecimation                               === */
////////////////////////////////////////////////////////////////////////////////
/* ========================================================================== */
////////////////////////////////////////////////////////////////////////////////

vtkCxxRevisionMacro(vtkUnstructuredGridQuadricDecimation, "$Revision: 1.4 $");
vtkStandardNewMacro(vtkUnstructuredGridQuadricDecimation);

void vtkUnstructuredGridQuadricDecimation::PrintSelf(ostream& os, vtkIndent indent)
  {
  os << indent << "Debug: " << (this->Debug ? "On\n" : "Off\n");
  os << indent << "Modified Time: " << this->GetMTime() << "\n";  this->Superclass::PrintSelf(os,indent);
  this->Superclass::PrintSelf(os, indent);
  os << indent << "Target Reduction: " << this->TargetReduction << "\n";
  os << indent << "Number of Tets to Output: " << this->NumberOfTetsOutput << "\n";
  os << indent << "Number of Edges to Decimate: " << this->NumberOfEdgesToDecimate << "\n";
  os << indent << "Number of Candidates Per Set: " << this->NumberOfCandidates << "\n";
  os << indent << "AutoAddCandidates: " << this->AutoAddCandidates << "\n";
  os << indent << "AutoAddCandidatesThreshold: " << this->AutoAddCandidatesThreshold << "\n";
  os << indent << "Boundary Weight: " << this->BoundaryWeight << "\n";
  }

vtkUnstructuredGridQuadricDecimation::vtkUnstructuredGridQuadricDecimation()
  {
  TargetReduction = 1.0;  
  NumberOfTetsOutput = 0;
  NumberOfEdgesToDecimate = 0;
  NumberOfCandidates = 8;
  AutoAddCandidates = 1;
  AutoAddCandidatesThreshold = 0.4;
  BoundaryWeight = 100.0;
  }

vtkUnstructuredGridQuadricDecimation::~vtkUnstructuredGridQuadricDecimation()
  {
  }


int vtkUnstructuredGridQuadricDecimation::RequestData(
  vtkInformation *vtkNotUsed(request),
  vtkInformationVector **inputVector,
  vtkInformationVector *outputVector)
  {
  // get the info objects
  vtkInformation *inInfo = inputVector[0]->GetInformationObject(0);
  vtkInformation *outInfo = outputVector->GetInformationObject(0);

  // get the input and ouptut
  vtkUnstructuredGrid *input = vtkUnstructuredGrid::SafeDownCast(inInfo->Get(vtkDataObject::DATA_OBJECT()));
  vtkUnstructuredGrid *output = vtkUnstructuredGrid::SafeDownCast(outInfo->Get(vtkDataObject::DATA_OBJECT()));

  vtkUnstructuredGridQuadricDecimationTetMesh myMesh;
  myMesh.doublingRatio = this->AutoAddCandidatesThreshold;
  myMesh.noDoubling = !this->AutoAddCandidates;
  myMesh.boundaryWeight = this->BoundaryWeight;
  input->Update();
  myMesh.loadVtkUnstructuredGrid((vtkUnstructuredGrid*)(input));

  myMesh.buildFullMesh();

  int desiredTets = this->NumberOfTetsOutput;
  if (desiredTets==0)
    desiredTets = (int)((1 - this->TargetReduction) * myMesh.tCount);
  myMesh.simplify(this->NumberOfEdgesToDecimate, desiredTets);
  myMesh.saveVtkUnstructuredGrid(output);
  return 1;
  }
