#ifndef __vtkUnstructuredGridQuadricDecimation_h
#define __vtkUnstructuredGridQuadricDecimation_h

#include "vtkUnstructuredGridAlgorithm.h"

class VTK_GRAPHICS_EXPORT vtkUnstructuredGridQuadricDecimation : public vtkUnstructuredGridAlgorithm 
{
public:
  vtkTypeRevisionMacro(vtkUnstructuredGridQuadricDecimation, vtkUnstructuredGridAlgorithm);
  void PrintSelf(ostream& os, vtkIndent indent);
  static vtkUnstructuredGridQuadricDecimation *New();

  // The following 3 parameters will control the process of simplification in the priority:
  // NumberOfEdgesToDecimate, NumberOfTetsOutput, TargetReduction.
  // If NumberOfEdgesToDeceimate is 0, NumberOfTetsOutput will be considered.
  // And if NumbersOfTetsOutput is also 0, then TargetReduction will control the output.
  
  // Description:
  // Set/Get the desired reduction (expressed as a fraction of the original
  // number of tetrehedra)
  vtkSetMacro(TargetReduction, double);
  vtkGetMacro(TargetReduction, double);
  
  // Description:
  // Set/Get the desired number of tetrahedra to be outputed
  vtkSetMacro(NumberOfTetsOutput, int);
  vtkGetMacro(NumberOfTetsOutput, int);

  // Description:
  // Set/Get the desired number of edge to collapse
  vtkSetMacro(NumberOfEdgesToDecimate, int);
  vtkGetMacro(NumberOfEdgesToDecimate, int);

  // Description:
  // Set/Get the number of candidates selected for each randomized set before
  // performing an edge collapse. Increasing this number can help producing
  // higher quality output but it will be slower. Default is 8.
  vtkSetMacro(NumberOfCandidates, int);
  vtkGetMacro(NumberOfCandidates, int);
  
  // Description:
  // Enable(1)/Disable(0) the feature of temporarily doubling the number of
  // candidates for each randomized set if the quadric error was significantly
  // increased over the last edge collapse, i.e. if the ratio between the error
  // difference and the last error is over some threshold. Basically, we are
  // trying to make up for cases when random selection returns so many 'bad' edges.
  // By doing this we can achieve a higher quality output with much less time
  // than just double the NumberOfCandidates. Default is Enabled(1)
  vtkSetMacro(AutoAddCandidates, int);
  vtkGetMacro(AutoAddCandidates, int);
  
  // Description:
  // Set/Get the threshold that decides when to double the set size. Default is 0.4.
  vtkSetMacro(AutoAddCandidatesThreshold, double);
  vtkGetMacro(AutoAddCandidatesThreshold, double);
  
  // Description:
  // Set/Get the weight of the boundary on the quadric metrics. The larger
  // the number, the better the boundary is preserved.
  vtkSetMacro(BoundaryWeight, double);
  vtkGetMacro(BoundaryWeight, double);

protected:
  vtkUnstructuredGridQuadricDecimation();
  ~vtkUnstructuredGridQuadricDecimation();

  int RequestData(vtkInformation *, vtkInformationVector **, vtkInformationVector *);

  int NumberOfTetsOutput;
  int NumberOfEdgesToDecimate;
  int NumberOfCandidates;
  int AutoAddCandidates;
  
  double TargetReduction;
  double AutoAddCandidatesThreshold;
  double BoundaryWeight;

private:
  vtkUnstructuredGridQuadricDecimation(const vtkUnstructuredGridQuadricDecimation&);  // Not implemented.
  void operator=(const vtkUnstructuredGridQuadricDecimation&);  // Not implemented.

}; 

#endif
