#include "vtkCORIEUnstructuredGridReader.h"
#include "vtkCellArray.h"
#include "vtkFieldData.h"
#include "vtkObjectFactory.h"
#include "vtkPolyData.h"
#include "vtkFloatArray.h"
#include "vtkUnstructuredGrid.h"
#include "vtkProperty2D.h"
#include "vtkPointData.h"
#include "vtkPolyDataMapper.h"
#include "vtkCollection.h"
#include "vtkContourGrid.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <time.h>
#include "elio.h"
#include "elio.cpp"

////////////////////////////////////////////////////////////////////////////////
/* ========================================================================== */
////////////////////////////////////////////////////////////////////////////////
/* === vtkUnstructuredGridReader                                          === */
////////////////////////////////////////////////////////////////////////////////
/* ========================================================================== */
////////////////////////////////////////////////////////////////////////////////

vtkCxxRevisionMacro(vtkCORIEUnstructuredGridReader, "$Revision: 1.1.1.1 $");
vtkStandardNewMacro(vtkCORIEUnstructuredGridReader);

// IMPORTED FROM Elcirc
vtkUnstructuredGrid *vtkCORIEUnstructuredGridReader::ReadTimeStep(vtkUnstructuredGrid *ugrid, char *fileName) {
  int i, j;
  FILE *fp = NULL;
  ElcircHeader m_h;
  ElioGetHeader(fileName, &m_h);
  //  ElioPrintHeader(&m_h);
  if (TimeStep>=m_h.nsteps) TimeStep = m_h.nsteps-1;
  float pt[4];
  int con[8];
  vtkPoints *points = vtkPoints::New();
  // Create the computational grid
  for (j = 0; j < m_h.nvrt; j++) {
	for (i = 0; i < m_h.np; i++) {
	  // node number is (i + j * nnodes)	
	  pt[0] = m_h.x[i];
	  pt[1] = m_h.y[i];
	  if (m_h.bi[i] == j) { // bottom index, set node Z values to depth of grid
		pt[2] = -m_h.d[i];
	  } else {
		pt[2] = -(m_h.zmsl - m_h.zcor[j]);
	  }
	  // set the vertical scale
	  pt[2] = pt[2] * this->VerticalScale + this->VerticalShift;
	  points->InsertPoint(i + j * m_h.np, pt);
	}
  }

  ElcircTimeStep t;
  ElioAllocateTimeStep(&m_h, &t);
  float *data = new float[m_h.np * m_h.ivs * m_h.nvrt];
  fp = fopen(fileName, "rb");
  if (ElioGetTimeStep(fp, TimeStep, &m_h, &t)) {
	printf("Leaving early\n");
  }
  //  printf("Time = %f, %d\n", t.t, t.it);
  vtkFloatArray *s = vtkFloatArray::New();  
  s->SetNumberOfValues(m_h.nvrt*m_h.np);  
  vtkFloatArray *vect = vtkFloatArray::New();
  if (m_h.ivs==1)
	ElioMakeScalarsOld(&m_h, t.d, data);
  else {
	ElioMakeVectorsOld(&m_h, t.d, data);
	vect->SetNumberOfValues(m_h.nvrt*m_h.np*3);
	vect->SetNumberOfComponents(3);
  }
  for (int k = 0; k < m_h.nvrt; k++) {
	for (i = 0; i < m_h.np; i++) {
	  if (m_h.ivs==1)
		s->SetValue(i + k*m_h.np, fabs(data[i*m_h.nvrt+k]));
	  else {
		float a = data[2*(i*m_h.nvrt+k)];
		vect->SetValue(3*(i + k*m_h.np), a);
		float sum = a*a;
		a = data[2*(i*m_h.nvrt+k)+1];
		vect->SetValue(3*(i + k*m_h.np)+1, a);
		sum += a*a;
		s->SetValue(i + k*m_h.np, sqrt(sum));
		vect->SetValue(3*(i + k*m_h.np)+2, 0.0);
	  }
	}
  }
  // Set the scalar range to 0..1

  delete[]t.d;
  delete[]data;

  // Transform scalar values to [0...)
  //vtkFloatArray *tmp = s;
  //s = vtkFloatArray::New();
  //double range[2];
  //tmp->GetRange(range);
  //s->SetNumberOfValues(tmp->GetNumberOfTuples());
  //for (i=s->GetNumberOfTuples()-1; i>=0; --i) {
  //	s->SetValue(i, (tmp->GetValue(i) - range[0]));
  //}
  //tmp->Delete();

  // Create grid
  int n1, n2, n3, n4;
  int s1, s2, s3, s4;
  int b1, b2, b3, b4;
  double reflevel;

  // if we only have 1 layer?
  if (m_h.nvrt==1) {
	ugrid->Allocate(m_h.ne);
	j = 0;
	reflevel = m_h.zmsl - m_h.zcor[j];
	for (i = 0; i < m_h.ne; i++) {
	  // element number is i + j * (maxlevels - 1)
	  if (m_h.etype[i] == 3) {
		n1 = m_h.icon[0][i];
		n2 = m_h.icon[1][i];
		n3 = m_h.icon[2][i];
		con[0] = n1 + j * m_h.np;
		con[1] = n2 + j * m_h.np;
		con[2] = n3 + j * m_h.np;
		for (int k = 0; k < 3; k++) {
		  if (s->GetValue(con[k]) < 0.0) {
			goto noel1;
		  }
		}
		ugrid->InsertNextCell(VTK_TRIANGLE, 3, con);
	  } else {
		n1 = m_h.icon[0][i];
		n2 = m_h.icon[1][i];
		n3 = m_h.icon[2][i];
		n4 = m_h.icon[3][i];
		con[0] = n1 + j * m_h.np;
		con[1] = n2 + j * m_h.np;
		con[2] = n3 + j * m_h.np;
		con[3] = n4 + j * m_h.np;

		for (int k = 0; k < 4; k++) {
		  if (s->GetValue(con[k])<0.0) {
			goto noel1;
		  }
		}
		ugrid->InsertNextCell(VTK_QUAD, 4, con);
	  }
	noel1:;
	}
  }
  else {	// we have more than 1 layer
	ugrid->Allocate((m_h.nvrt-1)*m_h.ne);
	if (this->FromLayer<0) this->FromLayer = 0;
	if (this->ToLayer>m_h.nvrt-1) this->ToLayer = m_h.nvrt-1;
	for (j = this->FromLayer; j < this->ToLayer; j++) {
	  reflevel = m_h.zmsl - m_h.zcor[j];
	  for (i = 0; i < m_h.ne; i++) {
		// element number is i + j * (maxlevels - 1)
		if (m_h.etype[i] == 3) {
		  n1 = m_h.icon[0][i];
		  n2 = m_h.icon[1][i];
		  n3 = m_h.icon[2][i];

		  s1 = t.surfind[n1];
		  s2 = t.surfind[n2];
		  s3 = t.surfind[n3];

		  b1 = m_h.bi[n1];
		  b2 = m_h.bi[n2];
		  b3 = m_h.bi[n3];

		  if (s1 < j+1 && s2 < j+1 && s3 < j+1) continue; // hack to filter out of domain elements.
		  if (b1 > j && b2 > j && b3 > j) continue; // hack to filter out of domain elements.
		  con[0] = n3 + j * m_h.np;
		  con[1] = n2 + j * m_h.np;
		  con[2] = n1 + j * m_h.np;
		  con[3] = n3 + (j + 1) * m_h.np;
		  con[4] = n2 + (j + 1) * m_h.np;
		  con[5] = n1 + (j + 1) * m_h.np;
		  for (int k = 0; k < 6; k++) {
			if (s->GetValue(con[k]) < 0.0) {
			  goto noel;
			}
		  }		  
		  ugrid->InsertNextCell(VTK_WEDGE, 6, con);
		} else {
		  n1 = m_h.icon[0][i];
		  n2 = m_h.icon[1][i];
		  n3 = m_h.icon[2][i];
		  n4 = m_h.icon[3][i];

		  s1 = t.surfind[n1];
		  s2 = t.surfind[n2];
		  s3 = t.surfind[n3];
		  s4 = t.surfind[n4];

		  b1 = m_h.bi[n1];
		  b2 = m_h.bi[n2];
		  b3 = m_h.bi[n3];
		  b4 = m_h.bi[n3];

		  if (s1 < j+1 && s2 < j+1 && s3 < j+1 && s4 < j + 1) continue; // hack to filter out out of domain elements.
		  if (b1 > j && b2 > j && b3 > j && b4 > j) continue; // hack to filter out of domain elements.
		  con[0] = n1 + j * m_h.np;
		  con[1] = n2 + j * m_h.np;
		  con[2] = n3 + j * m_h.np;
		  con[3] = n4 + j * m_h.np;
		  con[4] = n1 + (j + 1) * m_h.np;
		  con[5] = n2 + (j + 1) * m_h.np;
		  con[6] = n3 + (j + 1) * m_h.np;
		  con[7] = n4 + (j + 1) * m_h.np;
		  for (int k = 0; k < 8; k++) {
			if (s->GetValue(con[k]) < 0.0) {
			  goto noel;
			}
		  }
		  ugrid->InsertNextCell(VTK_HEXAHEDRON, 8, con);
		}
	  }
	noel:;
	}
  }
  delete[]t.surfind;
  fclose(fp);

  ugrid->SetPoints(points);
  points->Delete();
  s->SetName("scalars");
  ugrid->GetPointData()->AddArray(s);
  ugrid->GetPointData()->SetScalars(s);
  if (m_h.ivs>1)
	ugrid->GetPointData()->SetVectors(vect);
  s->Delete();
  return ugrid;
}
  
vtkUnstructuredGrid *vtkCORIEUnstructuredGridReader::ReadUGrid(vtkUnstructuredGrid *ugrid, const char *fname) {
  char buf[1024];
  int con[8];
  int i, nmel, nmnp, itmp, eltype, i1, i2, i3, i4;
  double x, y, d;
  FILE *fp = fopen(fname, "r");
  if (fp == 0) {
	return 0;
  }

  if (fgets(buf, 255, fp) == 0) {
	fclose(fp);
	return 0;
  }
  if (fgets(buf, 255, fp) == 0) {
	fclose(fp);
	return 0;
  }
  sscanf(buf, "%d %d", &nmel, &nmnp);
  vtkPoints *points = vtkPoints::New();
  vtkFloatArray *s = vtkFloatArray::New();
  points->SetNumberOfPoints(nmnp);
  s->SetNumberOfValues(nmnp);
  s->SetNumberOfComponents(3);
  ugrid->Allocate(nmel);

  for (i = 0; i < nmnp; i++) {
	if (fgets(buf, 255, fp) != 0) {
	  sscanf(buf, "%d %lf %lf %lf", &itmp, &x, &y, &d);
	  points->SetPoint(i, x, y, -d * this->VerticalScale + this->VerticalShift);
	  s->SetValue(i, -d);
	} else {
	  points->Delete();
	  s->Delete();
	  fclose(fp);
	  return 0;
	}
  }
		
  // Transform scalar values to [0...)
  vtkFloatArray *tmp = s;
  s = vtkFloatArray::New();
  double range[2];
  tmp->GetRange(range);
  s->SetNumberOfValues(tmp->GetNumberOfTuples());
  for (i=s->GetNumberOfTuples()-1; i>=0; --i) {
	s->SetValue(i, (tmp->GetValue(i) - range[0]));
  }
  tmp->Delete();

  for (i = 0; i < nmel; i++) {
	if (fgets(buf, 255, fp) != 0) {
	  sscanf(buf, "%d %d", &itmp, &eltype);
	  switch (eltype) {
	  case 3:
		sscanf(buf, "%d %d %d %d %d", &itmp, &eltype, &i1, &i2, &i3);
		con[0] = i1 - 1;
		con[1] = i2 - 1;
		con[2] = i3 - 1;
		ugrid->InsertNextCell(VTK_TRIANGLE, 3, con);
		break;
	  case 4:
		sscanf(buf, "%d %d %d %d %d %d", &itmp, &eltype, &i1, &i2, &i3, &i4);
		con[0] = i1 - 1;
		con[1] = i2 - 1;
		con[2] = i3 - 1;
		con[3] = i4 - 1;
		ugrid->InsertNextCell(VTK_QUAD, 4, con);
		break;
	  case 5:
		sscanf(buf, "%d %d %d %d %d %d", &itmp, &eltype, &i1, &i2, &i3, &i4);
		con[0] = i1 - 1;
		con[1] = i2 - 1;
		con[2] = i3 - 1;
		con[3] = i4 - 1;
		ugrid->InsertNextCell(VTK_TETRA, 4, con);
		break;
	  }
	} else {
	  points->Delete();
	  s->Delete();
	  ugrid->Delete();
	  fclose(fp);
	  return 0;
	}
  }
  fclose(fp);

  ugrid->SetPoints(points);
  points->Delete();
  s->SetName("scalars");
  ugrid->GetPointData()->AddArray(s);
  ugrid->GetPointData()->SetScalars(s);
  s->Delete();
  return ugrid;
}


void vtkCORIEUnstructuredGridReader::PrintSelf(ostream& os, vtkIndent indent) {
  this->Superclass::PrintSelf(os,indent);
}

vtkCORIEUnstructuredGridReader::vtkCORIEUnstructuredGridReader() {
	this->VerticalScale = 45.0;
	this->VerticalShift = 0.0;
	this->TimeStep = 0;
	this->FromLayer = 0;
	this->ToLayer = 10000;
}

vtkCORIEUnstructuredGridReader::~vtkCORIEUnstructuredGridReader() {
}

int vtkCORIEUnstructuredGridReader::RequestData(
  vtkInformation *,
  vtkInformationVector **,
	vtkInformationVector *outputVector) {

	if (strlen(FileName)<4 || strcmp(&this->FileName[strlen(this->FileName)-3], "gr3")) {
		this->ReadTimeStep(this->GetOutput(), this->FileName);
	}
	else
		this->ReadUGrid(this->GetOutput(), this->FileName);
	return 1;
}
